#------------------------------------------------------------------
#---
#--- ALLPROSE
#--- Copyright (C) Ralf Hemmecke (ralf@hemmecke.de)
#--- http://www.hemmecke.de/aldor
#---
#------------------------------------------------------------------

my ($inCodeChunk)=0;
my ($inVerbatim)='';
my ($inSnippet)='';
my ($initialSnippetSpaces)=-1;
my (@snippetContents)=();
while(<>) {
    if (/^<<.*>>=/) {$inCodeChunk=1; print; next;}
    if ($inCodeChunk) {
        if (/^\@($|\s)/) {$inCodeChunk=0;}
        print; next;
    }
    if (/^\s*\\begin{([^}]*verbatim[^}]*)}/) {$inVerbatim=$1; print; next;}
    if ($inVerbatim ne '') {#in verbatim mode
        if (/^\s*\\end{$inVerbatim}/) {$inVerbatim='';}
        print; next;
    }
    if (/^\s*\\begin{ad(usage|snippet)}/) {$inSnippet=$1; print; next;}
    if ($inSnippet ne '') {#in snippet mode
        if (/^\s*\\end{ad$inSnippet}/) {
            $inSnippet='';
            $initialSnippetSpaces = -1;
            print join("\\mbox{}\\\\\n", @snippetContents) . "\n";
            @snippetContents=();
            print;
            next;
        }
        if ($initialSnippetSpaces < 0) {
            /^(\s*)/;
            $initialSnippetSpaces = length($1);
        }
        chomp; # remove trailing newline;
        s/^\s{0,$initialSnippetSpaces}//; # remove leading spaces
        push @snippetContents, &adtranslateAdsnippetText($_);
        next;
    }
    chomp; #remove trailing newline;
    print &adtranslateLine($_), "\n";
}
sub adidentity {my($s)=@_; return $s;}
sub adtranslateLine {
    my ($after) = @_; # we assume that there is no newline at the end!
    my ($before, $code, $ret, $delim) = ('','','','');

    while ($after =~ /\\adcode(.)/) {
        $before = $`;
        $delim = $1;
        $after = $'; #'
        if ($delim eq '{') {$delim = '}';}
        $pos = index($after, $delim);
        if ($pos == -1) {
            print STDERR "ERROR: aldordoc2tex.pl:\\adcode not closed.\n";
            print STDERR "$_[0]\n";
            die "";
        }
        $code = substr($after, 0, $pos);
        $after = substr($after, $pos+1);
        $ret .= &adtranslateInterfaceOnlyCommands($before, \&adidentity);
        $ret .= &adtranslateAdcodeArgument($code, $delim);
    }
    $ret .= &adtranslateInterfaceOnlyCommands($after, \&adidentity);
    return $ret;
}
sub adtranslateAdsnippetText {
    my ($text) = @_;
    return &adtranslateActiveCommands($text, \&escapeForVerb, '')
}
sub adtranslateAdcodeArgument {
    my ($code, $delim) = @_;
    return &adtranslateActiveCommands($code, \&escapeForVerb, $delim)
}
sub adtranslateActiveCommands {
    my ($str, $escape, @escargs) = @_;
    my ($ret) = '';
    my ($b, $o, $type, $name, $c);
    my ($indexkey, $indextext);
    my ($hyperlabelname, $hyperlabelsignature);
    my ($latextext, $signature);
    my ($texcommand);

    while ($str=~/(\\ad(this)?(type|name))/) {
        $texcommand = $1;
        $this = $2;
        $ret .= &{$escape}($`, @escargs);
        $str = "$'"; # must copy {
        if ($this ne '') { # \adthisname and \adthistype
            $ret .= $texcommand;
            if ($str =~ /^{}/) {
                $str = $';
                $ret .= '{}';
            }
        } elsif ($texcommand eq '\adtype') {
            if ($str =~ /^({[A-Z][a-zA-Z0-9]*[a-z][a-zA-Z0-9]*})/) {
                $ret .= "\\adtype$1";
                $str = $';
            } else {
                print STDERR "adtranslateActiveCommands:\n";
                print STDERR "\\adtype{TYPE} must match regular expression\n";
                print STDERR "\\\\adtype{[A-Z][a-zA-Z0-9]*[a-z][a-zA-Z0-9]*}\n";
                die "Found: $str";
            }
        } else { # The \adname case.
            $str = "$texcommand$str";
            ($translated, $str) = &adtranslateInterfaceOnlyCommand($str);
            $ret .= $translated;
        }
    }
    $ret .= &{$escape}($str, @escargs);
    return $ret;
}
sub adtranslateInterfaceOnlyCommands {
    my ($str, $escape, @escargs) = @_;
    my ($ret, $translated) = ('', '');
    while ($str=~/(\\ad(parameter|(define)?name))/) {
        $ret .= &{$escape}($`, @escargs);
        $str = "$1$'";
        ($translated, $str) = &adtranslateInterfaceOnlyCommand($str);
        $ret .= $translated;
    }
    $ret .= &{$escape}($str, @escargs);
    return $ret;
}
sub adtranslateInterfaceOnlyCommand {
    my ($str) = @_;
    my ($ret) = '';
    my ($b, $o, $type, $name, $c);
    my ($indexkey, $indextext);
    my ($hyperlabelname, $hyperlabelsignature);
    my ($latextext, $signature);
    my ($texcommand);

    if ($str=~/^\\ad(parameter|(define)?name(\[[A-Z][a-zA-Z0-9]*[a-z][a-zA-Z0-9]*\])?){/) {
        $texcommand = $1;
        $str = "{$'"; # must copy {
        ($b, $o, $name, $c, $str) = &splitBalancedParentheses($str,'{','}');

        ($indexkey, $indextext,
         $hyperlabelname, $hyperlabelsignature,
         $latextext, $signature) = &splitNameAndSignatureArgument($name);

        if ($texcommand =~ /^name/) {$ret = 'use';} #\adinternalusename
        $ret = '\adinternal' . $ret . $texcommand;

        if ($texcommand eq 'parameter') {
            $ret .= "{$latextext}{" . &escapeForTTLatex($signature) . "}";
        } else {
            $ret .= "{$indexkey}{$indextext}"
                  . "{$hyperlabelname}{$hyperlabelsignature}"
                  . "{$latextext}";
        }
    } else {
        print STDERR "ERROR: adtranslateInterfaceOnlyCommand: " .
                     "No interface-only command found.\n";
        print STDERR "Current line is:\n";
        print STDERR "$str\n";
        die "";
    }
    return ($ret, $str);
}
sub splitBalancedParentheses {
    my ($s, $openparen, $closingparen)=@_;
    my ($before, $oparen, $text, $cparen) = ('', '', '', '');
    my ($after) = $s;
    if ($s =~ /($openparen)/) {
        $before = $`;
        $oparen = $1;
        ($text, $cparen, $after) =
          findClosingParenthesis($', $openparen, $closingparen);
    }
    return ($before, $oparen, $text, $cparen, $after);
}
sub findClosingParenthesis {
    my ($s, $openparen, $closingparen)=@_;
    my ($before, $open, $paren, $text, $close, $after);
    my ($closingParenFound)=0;
    my ($t, $p);
    $paren = '';
    $text = '';
    while ($s =~ /($openparen|$closingparen)/) {
        $text .= $`; # the part before the match
        $s = $'; # the part after the match
        $paren = $1;
        if ($paren =~ /$closingparen/) {$closingParenFound=1; last;}
        ($t, $p, $s) =
          findClosingParenthesis($s, $openparen, $closingparen);
        $text .= "$paren$t$p";
    }
    if ($closingParenFound==0) {die "Closing parenthesis not found";}
    return ($text, $paren, $s);
}
sub splitNameAndSignatureArgument {
    my($nameAndSignature)=@_;
    my($name);
    my($indexkey, $indextext);
    my($hyperlabelname, $hyperlabelsignature);
    my($latextext, $signature);
    if ($nameAndSignature =~ /\s*(.*?)\s*:\s*(.*)\s*/) {
        $name=$1;
        $signature=$2;
    } else {
        $name=$nameAndSignature;
        $signature='';
    }
    $indexkey            = &escapeForMakeindexKey($name);
    $indextext           = &escapeForMakeindexText($name);
    $hyperlabelname      = &escapeForHyperref($name);
    $hyperlabelsignature = &escapeForHyperref($signature);
    $latextext           = &escapeForTTLatex($name);
    ($indexkey, $indextext,
     $hyperlabelname, $hyperlabelsignature,
     $latextext, $signature);
}
sub escapeForVerb {
    my ($s, $delim) = @_;
    if ($s eq '') {return $s;}
    if ($delim eq '') {
        $delim = '&';
        $s = join("$delim\\verb|$delim|\verb$delim", split(/$delim/,$s));
    }
    return "\\verb$delim$s$delim";
}
sub escapeForHyperref {
    my($s)=@_;
    $s =~ s/\s//g;
    $s =~ s/(\(|{)/\[./g;
    $s =~ s/(\)|})/.\]/g;
    $s =~ s/%/PER/g;
    $s =~ s/_/.0/g;
    $s =~ s/\^/.1/g;
    $s =~ s/~/.2/g;
    $s =~ s/\\/.3/g;
    $s =~s/#/.4/g;
    return $s;
}
sub escapeForMakeindexText {
    my($s)=@_;
    $s = escapeForTTLatex($s);
    $s =~ s /!/\"!/g;
    return $s;
}
sub escapeForMakeindexKey {
    my($s)=@_;
    return &escapeForMakeindexText(&escapeForHyperref($s));
}
sub escapeForTTLatex {
    my($s)=@_;
    $s =~ s/\\/\\adbackslash /g;
    $s =~ s/{/\\{/g;
    $s =~ s/}/\\}/g;
    $s =~ s/_/\\_/g;
    $s =~ s/\^/\\^{}/g;
    $s =~ s/~/\\~{}/g;
    $s =~ s/%/\\%/g;
    $s =~ s/#/\\#/g;
    $s =~ s/&/\\&/g;
    return $s;
}
